@@ -2,12 +2,10 @@
 
 #include "../lib/common.h"
 #include "elf.h"
+#include "mold-git-hash.h"
 
-#include <atomic>
-#include <bitset>
 #include <cassert>
 #include <cstdint>
-#include <fstream>
 #include <functional>
 #include <iostream>
 #include <map>
@@ -15,13 +13,10 @@
 #include <mutex>
 #include <optional>
 #include <span>
-#include <sstream>
 #include <string>
 #include <string_view>
 #include <tbb/concurrent_hash_map.h>
-#include <tbb/concurrent_unordered_map.h>
 #include <tbb/concurrent_vector.h>
-#include <tbb/enumerable_thread_specific.h>
 #include <tbb/spin_mutex.h>
 #include <tbb/task_group.h>
 #include <type_traits>
@@ -55,7 +50,7 @@ struct ReaderContext;
 template <typename E>
 std::ostream &operator<<(std::ostream &out, const Symbol<E> &sym);
 
-std::string get_mold_version();
+extern std::string mold_version;
 
 //
 // Mergeable section fragments
@@ -71,9 +66,13 @@ struct __attribute__((aligned(4))) SectionFragment {
   }
 
   MergedSection<E> &output_section;
-  u32 offset = -1;
+  i64 offset = -1;
   Atomic<u8> p2align = 0;
   Atomic<bool> is_alive = false;
+
+  // True if this fragment must be placed within 2^32 bytes from the
+  // start of the output section.
+  Atomic<bool> is_32bit = false;
 };
 
 // Additional class members for dynamic symbols. Because most symbols
@@ -89,12 +88,11 @@ struct SymbolAux {
   i32 plt_idx = -1;
   i32 pltgot_idx = -1;
   i32 dynsym_idx = -1;
-  u32 djb_hash = 0;
-};
-
-template <>
-struct SymbolAux<PPC64V1> : SymbolAux<X86_64> {
   i32 opd_idx = -1;
+  u32 djb_hash = 0;
+
+  // For range extension thunks
+  std::vector<u64> thunk_addrs;
 };
 
 //
@@ -113,25 +111,60 @@ public:
   i64 size() const { return E::thunk_hdr_size + symbols.size() * E::thunk_size; }
   void copy_buf(Context<E> &ctx);
 
-  u64 get_addr(i64 idx) const {
-    return output_section.shdr.sh_addr + offset + E::thunk_hdr_size +
-           idx * E::thunk_size;
+  u64 get_addr() const {
+    return output_section.shdr.sh_addr + offset;
   }
 
-  static constexpr i64 alignment = 16;
+  u64 get_addr(i64 i) const {
+    return get_addr() + E::thunk_hdr_size + E::thunk_size * i;
+  }
 
   OutputSection<E> &output_section;
   i64 offset;
-  std::mutex mu;
   std::vector<Symbol<E> *> symbols;
+  std::string name;
 };
 
-struct ThunkRef {
-  static constexpr i64 MAX_SYM_IDX = (1 << 17) - 1;
+template <needs_thunk E>
+static consteval i64 get_branch_distance() {
+  // ARM64's branch has 26 bits immediate. The immediate is padded with
+  // implicit two-bit zeros because all instructions are 4 bytes aligned
+  // and therefore the least two bits are always zero. So the branch
+  // operand is effectively 28 bits long. That means the branch range is
+  // [-2^27, 2^27) or PC ± 128 MiB.
+  if (is_arm64<E>)
+    return 1 << 27;
 
-  i32 thunk_idx : 14 = -1;
-  i32 sym_idx : 18 = -1;
-};
+  // ARM32's Thumb branch has 24 bits immediate, and the instructions are
+  // aligned to 2, so it's effectively 25 bits. It's [-2^24, 2^24) or PC ±
+  // 16 MiB.
+  //
+  // ARM32's non-Thumb branches have twice longer range than its Thumb
+  // counterparts, but we conservatively use the Thumb's limitation.
+  if (is_arm32<E>)
+    return 1 << 24;
+
+  // PPC's branch has 24 bits immediate, and the instructions are aligned
+  // to 4, therefore the reach is [-2^25, 2^25) or PC ± 32 MiB.
+  assert(is_ppc<E>);
+  return 1 << 25;
+}
+
+// The maximum distance of branch instructions used for function calls.
+//
+// The exact origin for computing a destination varies slightly depending
+// on the target architecture. For example, ARM32's B instruction jumps to
+// the branch's address + immediate + 4 (i.e., B with offset 0 jumps to
+// the next instruction), while RISC-V has no such implicit bias. Here, we
+// subtract 32 as a safety margin that is large enough for all targets.
+template <needs_thunk E>
+static constexpr i64 branch_distance = get_branch_distance<E>() - 32;
+
+template <needs_thunk E>
+void remove_redundant_thunks(Context<E> &ctx);
+
+template <needs_thunk E>
+void gather_thunk_addresses(Context<E> &ctx);
 
 //
 // input-sections.cc
@@ -234,14 +267,29 @@ struct FdeRecord {
 template <typename E>
 struct InputSectionExtras {};
 
-template <needs_thunk E>
+template <is_arm32 E>
 struct InputSectionExtras<E> {
-  std::vector<ThunkRef> thunk_refs;
+  InputSection<E> *exidx = nullptr;
 };
 
+struct RelocDelta {
+  u64 offset;
+  i64 delta;
+};
+
+// RISC-V and LoongArch support code-shrinking linker relaxation.
+//
+// r_deltas is used to manage the locations where instructions are removed
+// from a section. r_deltas is sorted by offset. Each RelocDelta indicates
+// that the contents at and after `offset` and up to the next RelocDelta
+// offset need to be shifted towards the beginning of the section by
+// `delta` bytes when copying section contents to the output buffer.
+//
+// Since code-shrinking relaxation never bloats section contents, `delta`
+// increases monotonically within the vector as well.
 template <typename E> requires is_riscv<E> || is_loongarch<E>
 struct InputSectionExtras<E> {
-  std::vector<i32> r_deltas;
+  std::vector<RelocDelta> r_deltas;
 };
 
 // InputSection represents a section in an input object file.
@@ -251,7 +299,7 @@ public:
   InputSection(Context<E> &ctx, ObjectFile<E> &file, i64 shndx);
 
   void uncompress(Context<E> &ctx);
-  void copy_contents(Context<E> &ctx, u8 *buf);
+  void copy_contents_to(Context<E> &ctx, u8 *buf, i64 sz);
   void scan_relocations(Context<E> &ctx);
   void write_to(Context<E> &ctx, u8 *buf);
   void apply_reloc_alloc(Context<E> &ctx, u8 *base);
@@ -268,6 +316,7 @@ public:
   bool is_relr_reloc(Context<E> &ctx, const ElfRel<E> &rel) const;
   bool icf_removed() const;
   bool record_undef_error(Context<E> &ctx, const ElfRel<E> &rel);
+  void check_range(Context<E> &ctx, i64 i, i64 val, i64 lo, i64 hi);
 
   std::pair<SectionFragment<E> *, i64>
   get_fragment(Context<E> &ctx, const ElfRel<E> &rel);
@@ -324,8 +373,6 @@ private:
   void apply_toc_rel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel,
                      u8 *loc, u64 S, i64 A, u64 P, ElfRel<E> **dynrel);
 
-  u64 get_thunk_addr(i64 idx) requires needs_thunk<E>;
-
   std::optional<u64> get_tombstone(Symbol<E> &sym, SectionFragment<E> *frag);
 };
 
@@ -397,7 +444,7 @@ public:
   virtual i64 get_reldyn_size(Context<E> &ctx) const { return 0; }
   virtual void construct_relr(Context<E> &ctx) {}
   virtual void copy_buf(Context<E> &ctx) {}
-  virtual void write_to(Context<E> &ctx, u8 *buf, ElfRel<E> *rel) { unreachable(); }
+  virtual void write_to(Context<E> &ctx, u8 *buf) { unreachable(); }
   virtual void update_shdr(Context<E> &ctx) {}
 
   std::string_view name;
@@ -413,8 +460,8 @@ public:
   // For example, range extension thunks adds function_name@thunk
   // symbol for each thunk entry. The following members are used
   // for such synthesizing symbols.
-  virtual void compute_symtab_size(Context<E> &ctx) {};
-  virtual void populate_symtab(Context<E> &ctx) {};
+  virtual void compute_symtab_size(Context<E> &ctx) {}
+  virtual void populate_symtab(Context<E> &ctx) {}
 
   i64 local_symtab_idx = 0;
   i64 num_local_symtab = 0;
@@ -522,7 +569,7 @@ public:
   i64 get_reldyn_size(Context<E> &ctx) const override;
   void construct_relr(Context<E> &ctx) override;
   void copy_buf(Context<E> &ctx) override;
-  void write_to(Context<E> &ctx, u8 *buf, ElfRel<E> *rel) override;
+  void write_to(Context<E> &ctx, u8 *buf) override;
 
   void compute_symtab_size(Context<E> &ctx) override;
   void populate_symtab(Context<E> &ctx) override;
@@ -535,6 +582,9 @@ public:
   std::unique_ptr<RelocSection<E>> reloc_sec;
   std::vector<AbsRel<E>> abs_rels;
   Atomic<u32> sh_flags;
+
+  // Used only by create_output_sections()
+  std::vector<std::vector<InputSection<E> *>> members_vec;
 };
 
 template <typename E>
@@ -548,9 +598,9 @@ public:
     this->shdr.sh_addralign = sizeof(Word<E>);
 
     // We always create a .got so that _GLOBAL_OFFSET_TABLE_ has
-    // something to point to. s390x psABI define GOT[1] as a
-    // reserved slot, so we allocate one more for them.
-    this->shdr.sh_size = (is_s390x<E> ? 2 : 1) * sizeof(Word<E>);
+    // something to point to. s390x psABI define GOT[1] and GOT[2]
+    // as reserved slots, so we allocate two more for them.
+    this->shdr.sh_size = (is_s390x<E> ? 3 : 1) * sizeof(Word<E>);
   }
 
   void add_got_symbol(Context<E> &ctx, Symbol<E> *sym);
@@ -666,7 +716,6 @@ public:
   }
 
   void update_shdr(Context<E> &ctx) override;
-  void sort(Context<E> &ctx);
 };
 
 template <typename E>
@@ -726,8 +775,6 @@ public:
   i64 find_string(std::string_view str);
   void copy_buf(Context<E> &ctx) override;
 
-  i64 dynsym_offset = -1;
-
 private:
   std::unordered_map<std::string_view, i64> strings;
 };
@@ -799,6 +846,7 @@ public:
   void copy_buf(Context<E> &ctx) override;
 
   std::vector<Symbol<E> *> symbols;
+  i64 dynstr_offset = -1;
 };
 
 template <typename E>
@@ -856,7 +904,7 @@ public:
   void resolve(Context<E> &ctx);
   void compute_section_size(Context<E> &ctx) override;
   void copy_buf(Context<E> &ctx) override;
-  void write_to(Context<E> &ctx, u8 *buf, ElfRel<E> *rel) override;
+  void write_to(Context<E> &ctx, u8 *buf) override;
   void print_stats(Context<E> &ctx);
 
   std::vector<MergeableSection<E> *> members;
@@ -1258,13 +1306,13 @@ public:
   std::string_view get_contents(i64 idx);
 
   MergedSection<E> &parent;
+  u8 p2align = 0;
+  std::unique_ptr<InputSection<E>> input_section;
   std::vector<SectionFragment<E> *> fragments;
 
 private:
-  std::unique_ptr<InputSection<E>> section;
   std::vector<u32> frag_offsets;
   std::vector<u32> hashes;
-  u8 p2align = 0;
 };
 
 // InputFile is the base class of ObjectFile and SharedFile.
@@ -1284,9 +1332,7 @@ public:
 
   std::string_view get_string(Context<E> &ctx, const ElfShdr<E> &shdr);
   std::string_view get_string(Context<E> &ctx, i64 idx);
-
-  ElfEhdr<E> &get_ehdr() { return *(ElfEhdr<E> *)mf->data; }
-  std::span<ElfPhdr<E>> get_phdrs();
+  u32 get_eflags() { return ((ElfEhdr<E> *)mf->data)->e_flags; }
 
   ElfShdr<E> *find_section(i64 type);
 
@@ -1296,6 +1342,7 @@ public:
   mark_live_objects(Context<E> &ctx,
                     std::function<void(InputFile<E> *)> feeder) = 0;
 
+  std::span<Symbol<E> *> get_local_syms();
   std::span<Symbol<E> *> get_global_syms();
   std::string_view get_source_name() const;
 
@@ -1308,10 +1355,11 @@ public:
   std::string filename;
   bool is_dso = false;
   i64 priority;
-  Atomic<bool> is_alive = false;
+  Atomic<bool> is_reachable = false;
   std::string_view shstrtab;
   std::string_view symbol_strtab;
 
+  bool as_needed = false;
   bool has_init_array = false;
   bool has_ctors = false;
 
@@ -1331,7 +1379,8 @@ protected:
   std::vector<Symbol<E>> frag_syms;
 };
 
-template <typename E> struct ObjectFileExtras {};
+template <typename E>
+struct ObjectFileExtras {};
 
 template <is_riscv E>
 struct ObjectFileExtras<E> {
@@ -1351,8 +1400,8 @@ class ObjectFile : public InputFile<E> {
 public:
   ObjectFile() = default;
 
-  ObjectFile(Context<E> &ctx, MappedFile *mf,
-             std::string archive_name, bool is_in_lib);
+  ObjectFile(Context<E> &ctx, MappedFile *mf, std::string archive_name)
+    : InputFile<E>(ctx, mf), archive_name(archive_name) {}
 
   void parse(Context<E> &ctx);
   void initialize_symbols(Context<E> &ctx);
@@ -1374,19 +1423,20 @@ public:
   std::string archive_name;
   std::vector<std::unique_ptr<InputSection<E>>> sections;
   std::vector<std::unique_ptr<MergeableSection<E>>> mergeable_sections;
-  bool is_in_lib = false;
   std::vector<ElfShdr<E>> elf_sections2;
   std::vector<CieRecord<E>> cies;
   std::vector<FdeRecord<E>> fdes;
-  BitVector has_symver;
+  std::vector<bool> has_symver;
   std::vector<ComdatGroupRef<E>> comdat_groups;
   std::vector<InputSection<E> *> eh_frame_sections;
+  std::vector<std::vector<ElfRel<E>>> decoded_crel;
   bool exclude_libs = false;
   std::map<u32, u32> gnu_properties;
   bool needs_executable_stack = false;
   bool is_lto_obj = false;
   bool is_gcc_offload_obj = false;
   bool is_rust_obj = false;
+  bool is_dwarf32 = false;
 
   i64 fde_idx = 0;
   i64 fde_offset = 0;
@@ -1402,6 +1452,7 @@ public:
 
   // For LTO
   std::vector<ElfSym<E>> lto_elf_syms;
+  std::vector<ComdatGroup *> lto_comdat_groups;
 
 private:
   void initialize_sections(Context<E> &ctx);
@@ -1433,6 +1484,7 @@ public:
   std::span<Symbol<E> *> get_symbols_at(Symbol<E> *sym);
   i64 get_alignment(Symbol<E> *sym);
   std::vector<std::string_view> get_dt_needed(Context<E> &ctx);
+  std::string_view get_dt_audit(Context<E> &ctx);
   bool is_readonly(Symbol<E> *sym);
 
   void mark_live_objects(Context<E> &ctx,
@@ -1443,6 +1495,7 @@ public:
 
   std::string soname;
   std::vector<std::string_view> version_strings;
+  std::vector<Symbol<E> *> symbols2;
   std::vector<ElfSym<E>> elf_syms2;
 
 private:
@@ -1540,11 +1593,20 @@ void lto_cleanup(Context<E> &ctx);
 // shrink-sections.cc
 //
 
+inline i64 get_removed_bytes(std::span<RelocDelta> deltas, i64 i) {
+  if (i == 0)
+    return deltas[i].delta;
+  return deltas[i].delta - deltas[i - 1].delta;
+}
+
 template <typename E>
 void shrink_sections(Context<E> &ctx);
 
 template <typename E>
-void shrink_section(Context<E> &ctx, InputSection<E> &isec, bool use_rvc);
+void shrink_section(Context<E> &ctx, InputSection<E> &isec);
+
+template <typename E>
+i64 get_r_delta(InputSection<E> &isec, u64 offset);
 
 template <typename E>
 i64 compute_distance(Context<E> &ctx, Symbol<E> &sym,
@@ -1626,6 +1688,7 @@ template <typename E> void sort_init_fini(Context<E> &);
 template <typename E> void sort_ctor_dtor(Context<E> &);
 template <typename E> void fixup_ctors_in_init_array(Context<E> &);
 template <typename E> void shuffle_sections(Context<E> &);
+template <typename E> void add_dynamic_strings(Context<E> &);
 template <typename E> void compute_section_sizes(Context<E> &);
 template <typename E> void sort_output_sections(Context<E> &);
 template <typename E> void claim_unresolved_symbols(Context<E> &);
@@ -1633,6 +1696,7 @@ template <typename E> void scan_relocations(Context<E> &);
 template <typename E> void compute_imported_symbol_weakness(Context<E> &);
 template <typename E> void construct_relr(Context<E> &);
 template <typename E> void sort_dynsyms(Context<E> &);
+template <typename E> void sort_debug_info_sections(Context<E> &);
 template <typename E> void create_output_symtab(Context<E> &);
 template <typename E> void report_undef_errors(Context<E> &);
 template <typename E> void create_reloc_sections(Context<E> &);
@@ -1646,6 +1710,7 @@ template <typename E> void compute_section_headers(Context<E> &);
 template <typename E> i64 set_osec_offsets(Context<E> &);
 template <typename E> void fix_synthetic_symbols(Context<E> &);
 template <typename E> void compress_debug_sections(Context<E> &);
+template <typename E> void sort_reldyn(Context<E> &);
 template <typename E> void write_build_id(Context<E> &);
 template <typename E> void write_gnu_debuglink(Context<E> &);
 template <typename E> void write_separate_debug_file(Context<E> &ctx);
@@ -1662,28 +1727,34 @@ void rewrite_endbr(Context<X86_64> &ctx);
 // arch-arm32.cc
 //
 
-class Arm32ExidxSection : public Chunk<ARM32> {
+template <is_arm32 E>
+class Arm32ExidxSection : public Chunk<E> {
 public:
-  Arm32ExidxSection(OutputSection<ARM32> &osec) : output_section(osec) {
+  Arm32ExidxSection(OutputSection<E> &osec) : output_section(osec) {
     this->name = ".ARM.exidx";
     this->shdr.sh_type = SHT_ARM_EXIDX;
     this->shdr.sh_flags = SHF_ALLOC;
     this->shdr.sh_addralign = 4;
   }
 
-  void compute_section_size(Context<ARM32> &ctx) override;
-  void update_shdr(Context<ARM32> &ctx) override;
-  void remove_duplicate_entries(Context<ARM32> &ctx);
-  void copy_buf(Context<ARM32> &ctx) override;
+  void compute_section_size(Context<E> &ctx) override;
+  void update_shdr(Context<E> &ctx) override;
+  void remove_duplicate_entries(Context<E> &ctx);
+  void copy_buf(Context<E> &ctx) override;
 
 private:
-  std::vector<u8> get_contents(Context<ARM32> &ctx);
+  std::vector<u8> get_contents(Context<E> &ctx);
 
-  OutputSection<ARM32> &output_section;
+  OutputSection<E> &output_section;
 };
 
-template <> u64 get_eflags(Context<ARM32> &ctx);
-void create_arm_exidx_section(Context<ARM32> &ctx);
+template <is_arm32 E>
+u64 get_eflags(Context<E> &ctx);
+
+template <is_arm32 E>
+void create_arm_exidx_section(Context<E> &ctx);
+
+void arm32be_swap_bytes(Context<ARM32BE> &ctx);
 
 //
 // arch-riscv.cc
@@ -1833,9 +1904,9 @@ struct ContextExtras<E> {
   NotePropertySection<E> *note_property = nullptr;
 };
 
-template <>
-struct ContextExtras<ARM32> {
-  Arm32ExidxSection *exidx = nullptr;
+template <is_arm32 E>
+struct ContextExtras<E> {
+  Arm32ExidxSection<E> *exidx = nullptr;
 };
 
 template <is_riscv E>
@@ -1906,6 +1977,7 @@ struct Context {
     bool allow_multiple_definition = false;
     bool allow_shlib_undefined = true;
     bool apply_dynamic_relocs = true;
+    bool be8 = false;
     bool color_diagnostics = false;
     bool default_symver = false;
     bool demangle = true;
@@ -1986,8 +2058,11 @@ struct Context {
     i64 z_stack_size = 0;
     std::optional<Glob> unique;
     std::optional<u64> physical_image_base;
+    std::optional<std::vector<Symbol<E> *>> retain_symbols_file;
     std::string Map;
+    std::string audit;
     std::string chroot;
+    std::string depaudit;
     std::string dependency_file;
     std::string directory;
     std::string dynamic_linker;
@@ -1999,9 +2074,9 @@ struct Context {
     std::string soname;
     std::string sysroot;
     std::string_view emulation;
-    std::optional<std::vector<Symbol<E> *>> retain_symbols_file;
     std::unordered_map<std::string_view, u64> section_align;
     std::unordered_map<std::string_view, u64> section_start;
+    std::unordered_set<std::string_view> discard_section;
     std::unordered_set<std::string_view> ignore_ir_file;
     std::unordered_set<std::string_view> wrap;
     std::vector<SectionOrder> section_order;
@@ -2185,24 +2260,14 @@ enum {
   NEEDS_PPC_OPD   = 1 << 7, // for PPCv1
 };
 
-// A struct to hold target-dependent symbol members.
-template <typename E>
-struct SymbolExtras {};
-
-template <needs_thunk E>
-struct SymbolExtras<E> {
-  // For range extension thunks
-  i16 thunk_idx = -1;
-  i16 thunk_sym_idx = -1;
-};
-
 // Flags for Symbol<E>::get_addr()
 enum {
   NO_PLT = 1 << 0, // Request an address other than .plt
   NO_OPD = 1 << 1, // Request an address other than .opd (PPC64V1 only)
 };
 
-// Symbol class represents a defined symbol.
+// Symbol class represents a symbol. For each unique symbol name, we
+// create one instance of Symbol.
 //
 // A symbol has not only one but several different addresses if it
 // has PLT or GOT entries. This class provides various functions to
@@ -2255,6 +2320,8 @@ public:
   u32 get_djb_hash(Context<E> &ctx) const;
   void set_djb_hash(Context<E> &ctx, u32 hash);
 
+  u64 get_thunk_addr(Context<E> &ctx, u64 P) const requires needs_thunk<E>;
+
   bool is_absolute() const;
   bool is_relative() const { return !is_absolute(); }
   bool is_local(Context<E> &ctx) const;
@@ -2336,6 +2403,9 @@ public:
   bool is_traced : 1 = false;       // for --trace-symbol
   bool is_wrapped : 1 = false;      // for --wrap
 
+  // For symbols with default symbol version, e.g. foo@@VERSION.
+  bool is_versioned_default : 1 = false;
+
   // If a symbol can be resolved to a symbol in a different ELF file at
   // runtime, `is_imported` is true. If a symbol is a dynamic symbol and
   // can be used by other ELF file at runtime, `is_exported` is true.
@@ -2453,9 +2523,6 @@ public:
 
   // If true, we try to dmenagle the sybmol when printing.
   bool demangle : 1 = false;
-
-  // Target-dependent extra members.
-  [[no_unique_address]] SymbolExtras<E> extra;
 };
 
 template <typename E>
@@ -2465,6 +2532,9 @@ Symbol<E> *get_symbol(Context<E> &ctx, std::string_view key,
 template <typename E>
 Symbol<E> *get_symbol(Context<E> &ctx, std::string_view name);
 
+template <typename E>
+ComdatGroup *insert_comdat_group(Context<E> &ctx, std::string_view name);
+
 template <typename E>
 std::string_view demangle(const Symbol<E> &sym);
 
@@ -2536,7 +2606,11 @@ template <typename E>
 inline std::span<ElfRel<E>> InputSection<E>::get_rels(Context<E> &ctx) const {
   if (relsec_idx == -1)
     return {};
-  return file.template get_data<ElfRel<E>>(ctx, file.elf_sections[relsec_idx]);
+
+  ElfShdr<E> &shdr = file.elf_sections[relsec_idx];
+  if (shdr.sh_type == SHT_CREL)
+    return file.decoded_crel[relsec_idx];
+  return file.template get_data<ElfRel<E>>(ctx, shdr);
 }
 
 template <typename E>
@@ -2568,13 +2642,6 @@ InputSection<E>::get_fragment(Context<E> &ctx, const ElfRel<E> &rel) {
   return {p.first, p.second + get_addend(*this, rel)};
 }
 
-template <typename E>
-u64 InputSection<E>::get_thunk_addr(i64 idx) requires needs_thunk<E> {
-  ThunkRef ref = extra.thunk_refs[idx];
-  assert(ref.thunk_idx != -1);
-  return output_section->thunks[ref.thunk_idx]->get_addr(ref.sym_idx);
-}
-
 // Input object files may contain duplicate code for inline functions
 // and such. Linkers de-duplicate them at link-time. However, linkers
 // generaly don't remove debug info for de-duplicated functions because
@@ -2618,21 +2685,32 @@ inline bool InputSection<E>::icf_removed() const {
   return this->leader && this->leader != this;
 }
 
+template <typename E>
+inline void
+InputSection<E>::check_range(Context<E> &ctx, i64 i, i64 val, i64 lo, i64 hi) {
+  if (val < lo || hi <= val) {
+    const ElfRel<E> &rel = get_rels(ctx)[i];
+    Symbol<E> &sym = *file.symbols[rel.r_sym];
+    Error(ctx) << *this << ": relocation " << rel << " against "
+               << sym << " out of range: " << val << " is not in ["
+               << lo << ", " << hi << ")";
+  }
+}
+
 template <typename E>
 std::pair<SectionFragment<E> *, i64>
 MergeableSection<E>::get_fragment(i64 offset) {
-  std::span<u32> vec = frag_offsets;
-  auto it = std::upper_bound(vec.begin(), vec.end(), offset);
-  i64 idx = it - 1 - vec.begin();
-  return {fragments[idx], offset - vec[idx]};
+  auto it = ranges::upper_bound(frag_offsets, offset);
+  i64 idx = it - 1 - frag_offsets.begin();
+  return {fragments[idx], offset - frag_offsets[idx]};
 }
 
 template <typename E>
 std::string_view MergeableSection<E>::get_contents(i64 i) {
   i64 cur = frag_offsets[i];
   if (i == frag_offsets.size() - 1)
-    return section->contents.substr(cur);
-  return section->contents.substr(cur, frag_offsets[i + 1] - cur);
+    return input_section->contents.substr(cur);
+  return input_section->contents.substr(cur, frag_offsets[i + 1] - cur);
 }
 
 template <typename E>
@@ -2670,6 +2748,11 @@ inline std::string_view InputFile<E>::get_string(Context<E> &ctx, i64 idx) {
   return this->get_string(ctx, elf_sections[idx]);
 }
 
+template <typename E>
+inline std::span<Symbol<E> *> InputFile<E>::get_local_syms() {
+  return std::span<Symbol<E> *>(this->symbols).subspan(0, this->first_global);
+}
+
 template <typename E>
 inline std::span<Symbol<E> *> InputFile<E>::get_global_syms() {
   return std::span<Symbol<E> *>(this->symbols).subspan(this->first_global);
@@ -2953,6 +3036,22 @@ inline void Symbol<E>::set_djb_hash(Context<E> &ctx, u32 hash) {
   ctx.symbol_aux[aux_idx].djb_hash = hash;
 }
 
+template <typename E>
+u64
+Symbol<E>::get_thunk_addr(Context<E> &ctx, u64 P) const requires needs_thunk<E> {
+  std::span<u64> vec = ctx.symbol_aux[aux_idx].thunk_addrs;
+
+  u64 lo = (P < branch_distance<E>) ? 0 : P - branch_distance<E>;
+  auto it = ranges::lower_bound(vec, lo);
+  if (it == vec.end())
+    Fatal(ctx) << "range extension thunk out of range: " << *this;
+
+  i64 disp = *it - P;
+  if (disp < -branch_distance<E> || branch_distance<E> <= disp)
+    Fatal(ctx) << "range extension thunk out of range: " << *this;
+  return *it;
+}
+
 template <typename E>
 inline bool Symbol<E>::has_plt(Context<E> &ctx) const {
   return get_plt_idx(ctx) != -1 || get_pltgot_idx(ctx) != -1;
@@ -2960,6 +3059,11 @@ inline bool Symbol<E>::has_plt(Context<E> &ctx) const {
 
 template <typename E>
 inline bool Symbol<E>::is_absolute() const {
+  // An unresolved weak symbol acts as if it were an absolute address
+  // at address 0
+  if (is_remaining_undef_weak())
+    return true;
+
   return !is_imported && !get_frag() && !get_input_section() &&
          !get_output_section();
 }
@@ -3125,7 +3229,7 @@ std::string_view save_string(Context<E> &ctx, const std::string &str) {
   u8 *buf = new u8[str.size() + 1];
   memcpy(buf, str.data(), str.size());
   buf[str.size()] = '\0';
-  ctx.string_pool.push_back(std::unique_ptr<u8[]>(buf));
+  ctx.string_pool.emplace_back(buf);
   return {(char *)buf, str.size()};
 }
 
